//
// Copyright 2014 ADVANCED MICRO DEVICES, INC.  All Rights Reserved.
//
// AMD is granting you permission to use this software and documentation (if
// any) (collectively, the "Materials") pursuant to the terms and conditions
// of the Software License Agreement included with the Materials.  If you do
// not have a copy of the Software License Agreement, contact your AMD
// representative for a copy.
// You agree that you will not reverse engineer or decompile the Materials,
// in whole or in part, except as allowed by applicable law.
//
// WARRANTY DISCLAIMER: THE SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND.  AMD DISCLAIMS ALL WARRANTIES, EXPRESS, IMPLIED, OR STATUTORY,
// INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE, NON-INFRINGEMENT, THAT THE SOFTWARE
// WILL RUN UNINTERRUPTED OR ERROR-FREE OR WARRANTIES ARISING FROM CUSTOM OF
// TRADE OR COURSE OF USAGE.  THE ENTIRE RISK ASSOCIATED WITH THE USE OF THE
// SOFTWARE IS ASSUMED BY YOU.
// Some jurisdictions do not allow the exclusion of implied warranties, so
// the above exclusion may not apply to You. 
// 
// LIMITATION OF LIABILITY AND INDEMNIFICATION:  AMD AND ITS LICENSORS WILL
// NOT, UNDER ANY CIRCUMSTANCES BE LIABLE TO YOU FOR ANY PUNITIVE, DIRECT,
// INCIDENTAL, INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES ARISING FROM USE OF
// THE SOFTWARE OR THIS AGREEMENT EVEN IF AMD AND ITS LICENSORS HAVE BEEN
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.  
// In no event shall AMD's total liability to You for all damages, losses,
// and causes of action (whether in contract, tort (including negligence) or
// otherwise) exceed the amount of $100 USD.  You agree to defend, indemnify
// and hold harmless AMD and its licensors, and any of their directors,
// officers, employees, affiliates or agents from and against any and all
// loss, damage, liability and other expenses (including reasonable attorneys'
// fees), resulting from Your use of the Software or violation of the terms and
// conditions of this Agreement.  
//
// U.S. GOVERNMENT RESTRICTED RIGHTS: The Materials are provided with "RESTRICTED
// RIGHTS." Use, duplication, or disclosure by the Government is subject to the
// restrictions as set forth in FAR 52.227-14 and DFAR252.227-7013, et seq., or
// its successor.  Use of the Materials by the Government constitutes
// acknowledgement of AMD's proprietary rights in them.
// 
// EXPORT RESTRICTIONS: The Materials may be subject to export restrictions as
// stated in the Software License Agreement.
//
#ifndef __PARTICLE_SYSTEM_H__
#define __PARTICLE_SYSTEM_H__


#include "..\\DXUT\\Core\\DXUT.h"
#include "..\\AMD_SDK\\AMD_SDK.h"


// Implementation-agnostic particle system interface
struct IParticleSystem
{
	// The rendering technique to use
	enum Technique
	{
		Technique_Rasterize,
		Technique_Tiled,
		Technique_Overdraw,
		Technique_Max
	};

	// The coarse culling mode to use, if any
	enum CoarseCullingMode
	{
		CoarseCullingOff,
		CoarseCulling4x2,
		CoarseCulling8x8,
		CoarseCulling16x16,
		NumCoarseCullingModes
	};

	// Per-frame stats from the particle system
	struct Stats
	{
		int		m_MaxParticles;
		int		m_NumActiveParticles;
		int		m_NumDead;
	};

	enum Flags
	{
		PF_Sort = 1 << 0,				// Sort the particles
		PF_CheapLighting = 1 << 1,		// Perform minimal lighting on the particles to ease ALU load
		PF_NoLighting = 1 << 2,			// Do no lighting at all, just display the particle texture on the billboard
		PF_CullMaxZ = 1 << 3,			// Do per-tile MaxZ culling if applicable
		PF_Streaks = 1 << 4,			// Streak the particles based on velocity
		PF_UseGeometryShader = 1 << 5,	// Use the GS to do the billboarding, otherwise uses the VS for better performance
		PF_ScreenSpaceCulling = 1 << 6	// Do the tile culling in screen space to avoid potential false positives with frustum culling
	};

	// Per-emitter parameters
	struct EmitterParams
	{
		DirectX::XMVECTOR	m_Position;				// World position of the emitter
		DirectX::XMVECTOR	m_Velocity;				// Velocity of each particle from the emitter
		DirectX::XMVECTOR	m_PositionVariance;		// Variance in position of each particle
		int					m_NumToEmit;			// Number of particles to emit this frame
		float				m_ParticleLifeSpan;		// How long the particles should live
		float				m_StartSize;			// Size of particles at spawn time
		float				m_EndSize;				// Size of particle when they reach retirement age
		float				m_Mass;					// Mass of particle
		float				m_VelocityVariance;		// Variance in velocity of each particle
		int					m_TextureIndex;			// Index of the texture in the atlas
		bool				m_Streaks;				// Streak the particles in the direction of travel
	};

	// Create a GPU particle system. Add more factory functions to create other types of system eg CPU-updated system
	static IParticleSystem* CreateGPUSystem( AMD::ShaderCache& shadercache );

	virtual ~IParticleSystem() {}

	virtual void OnCreateDevice( ID3D11Device* pDevice, ID3D11DeviceContext* pImmediateContext ) = 0;
	virtual void OnResizedSwapChain( const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc ) = 0;
	virtual void OnReleasingSwapChain() = 0;
	virtual void OnDestroyDevice() = 0;

	// Completely resets the state of all particles. Handy for changing scenes etc
	virtual void Reset() = 0;

	// Render the system given a frame delta.
	virtual void Render( float frameTime, int flags, Technique technique, CoarseCullingMode coarseCullingMode, const EmitterParams* pEmitters, int nNumEmitters, ID3D11ShaderResourceView* depthSRV ) = 0;
	
	// Retrive the statistics about this frame's particles
	virtual const Stats& GetStats() const = 0;
};


#endif