//
// Copyright 2014 ADVANCED MICRO DEVICES, INC.  All Rights Reserved.
//
// AMD is granting you permission to use this software and documentation (if
// any) (collectively, the "Materials") pursuant to the terms and conditions
// of the Software License Agreement included with the Materials.  If you do
// not have a copy of the Software License Agreement, contact your AMD
// representative for a copy.
// You agree that you will not reverse engineer or decompile the Materials,
// in whole or in part, except as allowed by applicable law.
//
// WARRANTY DISCLAIMER: THE SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND.  AMD DISCLAIMS ALL WARRANTIES, EXPRESS, IMPLIED, OR STATUTORY,
// INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE, NON-INFRINGEMENT, THAT THE SOFTWARE
// WILL RUN UNINTERRUPTED OR ERROR-FREE OR WARRANTIES ARISING FROM CUSTOM OF
// TRADE OR COURSE OF USAGE.  THE ENTIRE RISK ASSOCIATED WITH THE USE OF THE
// SOFTWARE IS ASSUMED BY YOU.
// Some jurisdictions do not allow the exclusion of implied warranties, so
// the above exclusion may not apply to You. 
// 
// LIMITATION OF LIABILITY AND INDEMNIFICATION:  AMD AND ITS LICENSORS WILL
// NOT, UNDER ANY CIRCUMSTANCES BE LIABLE TO YOU FOR ANY PUNITIVE, DIRECT,
// INCIDENTAL, INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES ARISING FROM USE OF
// THE SOFTWARE OR THIS AGREEMENT EVEN IF AMD AND ITS LICENSORS HAVE BEEN
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.  
// In no event shall AMD's total liability to You for all damages, losses,
// and causes of action (whether in contract, tort (including negligence) or
// otherwise) exceed the amount of $100 USD.  You agree to defend, indemnify
// and hold harmless AMD and its licensors, and any of their directors,
// officers, employees, affiliates or agents from and against any and all
// loss, damage, liability and other expenses (including reasonable attorneys'
// fees), resulting from Your use of the Software or violation of the terms and
// conditions of this Agreement.  
//
// U.S. GOVERNMENT RESTRICTED RIGHTS: The Materials are provided with "RESTRICTED
// RIGHTS." Use, duplication, or disclosure by the Government is subject to the
// restrictions as set forth in FAR 52.227-14 and DFAR252.227-7013, et seq., or
// its successor.  Use of the Materials by the Government constitutes
// acknowledgement of AMD's proprietary rights in them.
// 
// EXPORT RESTRICTIONS: The Materials may be subject to export restrictions as
// stated in the Software License Agreement.
//
#include "ShaderConstants.h"


// Per-frame constant buffer
cbuffer PerFrameConstantBuffer : register( b0 )
{
	float4	g_StartColor[ NUM_EMITTERS ];
	float4	g_EndColor[ NUM_EMITTERS ];

	float4	g_EmitterLightingCenter[ NUM_EMITTERS ];

	matrix	g_mViewProjection;
	matrix	g_mViewProjInv;
	matrix	g_mView;
	matrix	g_mViewInv;
	matrix  g_mProjection;
	matrix  g_mProjectionInv;

	float4	g_EyePosition;
	float4	g_SunDirection;
	float4	g_SunColor;
	float4	g_AmbientColor;

	float4	g_SunDirectionVS;
	float4	pads2[ 3 ];

	float	g_fFrameTime;
	int		g_ScreenWidth;
	int		g_ScreenHeight;
	int		g_FrameIndex;
	
	float	g_AlphaThreshold;
	float	g_CollisionThickness;
	float	g_ElapsedTime;
	int		g_CollideParticles;

	int		g_ShowSleepingParticles;
	int		g_EnableSleepState;
	int		g_Pads[ 2 ];
};


// The number of dead particles in the system
cbuffer DeadListCount : register( b2 )
{
	uint	g_NumDeadParticles;
	uint3	DeadListCount_pad;
};


// The number of alive particles this frame
cbuffer ActiveListCount : register( b3 )
{
	uint	g_NumActiveParticles;
	uint3	ActiveListCount_pad;
};


// Tiling constants that are dependant on the screen resolution
cbuffer TilingConstantBuffer : register( b5 )
{
	uint g_NumTilesX;
	uint g_NumTilesY;
	uint g_NumCoarseCullingTilesX;
	uint g_NumCoarseCullingTilesY;

	uint g_NumCullingTilesPerCoarseTileX;
	uint g_NumCullingTilesPerCoarseTileY;
	uint g_TilingCBPads[ 2 ];
};


// Particle structures
// ===================

struct GPUParticlePartA
{
	float4	m_TintAndAlpha;			// The color and opacity
	float2	m_VelocityXY;			// View space velocity XY used for streak extrusion
	float	m_EmitterNdotL;			// The lighting term for the while emitter
	uint	m_EmitterProperties;	// The index of the emitter in 0-15 bits, 16-23 for atlas index, 24th bit is whether or not the emitter supports velocity-based streaks

	float	m_Rotation;				// The rotation angle
	uint	m_IsSleeping;			// Whether or not the particle is sleeping (ie, don't update position)
	uint	m_CollisionCount;		// Keep track of how many times the particle has collided
	float	m_pads[ 1 ];
};

struct GPUParticlePartB
{
	float3	m_Position;				// World space position
	float	m_Mass;					// Mass of particle

	float3	m_Velocity;				// World space velocity
	float	m_Lifespan;				// Lifespan of the particle.
	
	float	m_DistanceToEye;		// The distance from the particle to the eye
	float	m_Age;					// The current age counting down from lifespan to zero
	float	m_StartSize;			// The size at spawn time
	float	m_EndSize;				// The time at maximum age
};


uint GetEmitterIndex( uint emitterProperties )
{
	return emitterProperties & 0xffff;
}


float GetTextureOffset( uint emitterProperties )
{
	uint index = (emitterProperties & 0x00ff0000) >> 16;

	return (float)index / 2.0f;
}


bool IsStreakEmitter( uint emitterProperties )
{
	return ( emitterProperties >> 24 ) & 0x01 ? true : false;
}


uint WriteEmitterProperties( uint emitterIndex, uint textureIndex, bool isStreakEmitter )
{
	uint properties = emitterIndex & 0xffff;

	properties |= textureIndex << 16;

	if ( isStreakEmitter )
	{
		properties |= 1 << 24;
	}

	return properties;
}



// Function to calculate the streak radius in X and Y given the particles radius and velocity
float2 calcEllipsoidRadius( float radius, float2 viewSpaceVelocity )
{
	float minRadius = radius * max( 1.0, 0.1*length( viewSpaceVelocity ) );
	return float2( radius, minRadius );
}


// this creates the standard Hessian-normal-form plane equation from three points, 
// except it is simplified for the case where the first point is the origin
float3 CreatePlaneEquation( float3 b, float3 c )
{
	return normalize(cross( b, c ));
}


// point-plane distance, simplified for the case where 
// the plane passes through the origin
float GetSignedDistanceFromPlane( float3 p, float3 eqn )
{
    // dot( eqn.xyz, p.xyz ) + eqn.w, , except we know eqn.w is zero 
    // (see CreatePlaneEquation above)
    return dot( eqn, p );
}


// convert a point from post-projection space into view space
float3 ConvertProjToView( float4 p )
{
	p = mul( p, g_mProjectionInv );
	p /= p.w;
    
    return p.xyz;
}


// Declare the global samplers
SamplerState g_samWrapLinear		: register( s0 );
SamplerState g_samClampLinear		: register( s1 );

